from __future__ import annotations

import logging
import os
import typing as t
from datetime import datetime, timedelta, timezone

import aiohttp

from detaorm.paginator import RawPage, RawPaginator
from detaorm.query import Node
from detaorm.types import RAW_ITEM

if t.TYPE_CHECKING:
    from detaorm.base import Base
    from detaorm.update import Update

__all__ = ("Client", "RawPutItemsResponse")

_LOG = logging.getLogger(__name__)


def _with_ttl(
    dct: dict[str, object],
    expire_at: int | datetime | None,
    expire_in: int | timedelta | None,
) -> dict[str, object]:
    if expire_in and expire_at:
        raise ValueError("You cannot specify 'expire_at' and 'expire_in'.")

    if expire_in:
        if isinstance(expire_in, int):
            expire_in = timedelta(seconds=expire_in)

        expire_at = datetime.now(timezone.utc) + expire_in

    if isinstance(expire_at, datetime):
        expire_at = int(expire_at.timestamp())

    if expire_at:
        dct = dct.copy()
        dct["__expires"] = expire_at

    return dct


class RawPutItemsResponse:
    def __init__(
        self, processed: list[RAW_ITEM], failed: list[RAW_ITEM]
    ) -> None:
        self.processed = processed
        self.failed = failed


class Client:
    """A client for communicating with the DetaBase API.

    Args:
        project_key: The project key.
        bases: All of the models that will be used.
    """

    api_base_url = "https://database.deta.sh"
    api_version = "v1"

    def __init__(
        self, project_key: str | None = None, *, bases: t.Iterable[Base]
    ) -> None:
        for b in bases:
            b._client = self

        project_key = project_key or os.getenv("DETA_PROJECT_KEY")
        if not project_key:
            raise ValueError(
                "No project key provided. Please set a project key as an env "
                "variable, or pass it as an arg to Client."
            )

        self.project_key = project_key
        self.project_id = project_key.split("_")[0]
        self.__session: aiohttp.ClientSession | None = None

    @property
    def ready(self) -> bool:
        """Returns true if the client is ready to be used."""

        return not (self.__session is None or self.__session.closed)

    @property
    def _session(self) -> aiohttp.ClientSession:
        if not self.__session or self.__session.closed:
            raise RuntimeError("Client has not been opened yet.")
        return self.__session

    async def open(self) -> None:
        """Prepares the client to be used."""

        if self.ready:
            _LOG.warning("Client is already opened.")
            return

        self.__session = aiohttp.ClientSession(
            self.api_base_url,
            headers={
                "X-API-Key": self.project_key,
                "Content-Type": "application/json",
            },
            raise_for_status=True,
        )

    async def close(self) -> None:
        """Closes the aiohttp session."""

        if not self.ready:
            _LOG.warning("Client is not open.")
            return

        await self._session.close()
        self.__session = None

    # api methods
    async def put_items(
        self,
        base_name: str,
        items: t.Sequence[RAW_ITEM],
        *,
        expire_at: int | datetime | None = None,
        expire_in: int | timedelta | None = None,
    ) -> RawPutItemsResponse:
        """Stores multiple items in a single request.

        This request overwrites an item if the key already exists.

        Args:
            base_name: The name of the Base.
            items: A list of items to insert.
            expire_at: A Unix timestamp that these items should expire at.
            expire_in: The number of seconds until this item expires."""

        if len(items) > 25:
            _LOG.warning("Only 25 items can be inserted at a time.")

        items = [_with_ttl(i, expire_at, expire_in) for i in items]
        url = self._build_url(base_name, "items")
        resp = await self._session.put(url, json={"items": items})

        data: dict[str, dict[str, list[RAW_ITEM]]] = await resp.json()
        processed = data.get("processed", {}).get("items", [])
        failed = data.get("failed", {}).get("items", [])
        return RawPutItemsResponse(processed, failed)

    async def get_item(self, base_name: str, key: str) -> RAW_ITEM:
        """Get a stored item.

        Args:
            base_name: The name of the Base.
            key: The key of the item to get.

        Returns:
            A mapping of fields to values."""

        url = self._build_url(base_name, f"items/{key}")
        resp = await self._session.get(url)
        return t.cast(RAW_ITEM, await resp.json())

    async def delete_item(self, base_name: str, key: str) -> None:
        """Delete a stored item.

        Args:
            base_name: The name of the Base.
            key: The key of the item to delete."""

        url = self._build_url(base_name, f"items/{key}")
        await self._session.delete(url)

    async def insert_item(
        self,
        base_name: str,
        item: RAW_ITEM,
        *,
        expire_at: int | datetime | None = None,
        expire_in: int | timedelta | None = None,
    ) -> RAW_ITEM:
        """Creates a new item only if no item with the same key exists.

        Args:
            base_name: The name of the Base.
            item: The item to insert.
            expire_at: The time at which the item should expire (UTC timetamp).
            expire_in: The time until this item should expire."""

        json = _with_ttl({"item": item}, expire_at, expire_in)
        url = self._build_url(base_name, "items")
        resp = await self._session.post(url, json=json)
        return t.cast(RAW_ITEM, await resp.json())

    async def update_item(
        self,
        base_name: str,
        key: str,
        *updates: Update,
        set: dict[str, object] | None = None,
        increment: dict[str, int] | None = None,
        append: dict[str, list[object]] | None = None,
        prepend: dict[str, list[object]] | None = None,
        delete: list[str] | None = None,
    ) -> dict[str, object]:
        """Update an item.

        Args:
            base_name: The name of the Base.
            key: The key of the item to update.
            *updates: Update objects.
            set: A mapping of fields to set.
            increment: A mapping of fields to increment.
            append: A mapping of fields to append items to.
            prepend: A mapping of fields to prepend items to.
            delete: A list of fields to delete.

        Returns:
            The generated update query.
        """

        def join(left: object, right: object) -> object:
            if isinstance(left, t.Mapping):
                assert isinstance(right, t.Mapping)
                return {**left, **right}
            elif isinstance(left, t.Sequence):
                assert isinstance(right, t.Sequence)
                return [*left, *right]
            else:
                raise TypeError

        data = {
            "set": set or {},
            "increment": increment or {},
            "append": append or {},
            "prepend": prepend or {},
            "delete": delete or [],
        }

        for update in updates:
            if (val := data.get(update.field)) is not None:
                data[update.field] = join(val, update.payload)
            else:
                raise ValueError(f"Unexpected field {update.field}.")

        url = self._build_url(base_name, f"items/{key}")
        await self._session.patch(url, json=data)
        return data

    def query_items(
        self,
        base_name: str,
        query: Node | t.Sequence[t.Mapping[str, object]] | None = None,
        limit: int = 0,
        last: str | None = None,
    ) -> RawPaginator:
        """Query items.

        Args:
            base_name: The name of the Base.
            query: The query. Defaults to None.
            limit: The maximum page size. Defaults to 0 (no limit).
            last: The last key of the last page. Defaults to None.

        Returns:
            A RawPage. Use RawPage.items to see the items, and RawPage.next()
            to get the next page.
        """

        async def first_page(
            base_name: str,
            query: Node | t.Sequence[t.Mapping[str, object]] | None = None,
            limit: int = 0,
            last: str | None = None,
        ) -> RawPage:
            if isinstance(query, Node):
                query = query.deta_query()

            data: dict[str, object] = {}
            if query:
                data["query"] = query
            if limit:
                data["limit"] = limit
            if last:
                data["last"] = last

            url = self._build_url(base_name, "query")

            resp = await self._session.post(url, json=data)
            resp_data = await resp.json()
            return RawPage(
                client=self,
                base_name=base_name,
                query=query,
                size=resp_data["paging"]["size"],
                limit=limit,
                last=resp_data["paging"].get("last"),
                items=resp_data["items"],
            )

        return RawPaginator(first_page(base_name, query, limit, last))

    def _build_url(self, base_name: str, relative_path: str) -> str:
        return (
            f"/{self.api_version}/{self.project_id}/{base_name}/"
            f"{relative_path}"
        )



from detaorm.base import Base
from detaorm.client import Client
from detaorm.field import Field, NestedField
from detaorm.query import And, Node, Or, Query

__all__ = (
    "Base",
    "Client",
    "Field",
    "And",
    "Node",
    "Or",
    "Query",
    "NestedField",
)



from __future__ import annotations

from dataclasses import dataclass

__all___ = (
    "Updaate",
    "SetUpdate",
    "IncrementUpdate",
    "AppendUpdate",
    "PrependUpdate",
    "DeleteUpdate",
)


class Update:
    field: str
    payload: object


@dataclass
class SetUpdate:
    field = "set"
    payload: dict[str, object]


@dataclass
class IncrementUpdate:
    field = "increment"
    payload: dict[str, int]


@dataclass
class AppendUpdate:
    field = "append"
    payload: dict[str, list[object]]


@dataclass
class PrependUpdate:
    field = "prepend"
    payload: dict[str, list[object]]


@dataclass
class DeleteUpdate:
    field = "delete"
    payload: list[str]



from __future__ import annotations

import typing as t

__all__ = ("And", "Or", "Query", "Node")


class Node:
    @classmethod
    def _filter_nodes(cls, nodes: t.Sequence[Node]) -> t.Sequence[Node]:
        final: list[Node] = []
        for n in nodes:
            if isinstance(n, cls):
                final.extend(n.nodes)
            else:
                final.append(n)
        return final

    @property
    def nodes(self) -> t.Sequence[Node]:
        raise NotImplementedError

    def deta_query(self) -> list[dict[str, object]]:
        result: list[dict[str, object]] = []
        tree = self._flatten()
        if not isinstance(tree, Or):
            tree = Or(tree)
        for node in tree.nodes:
            assert isinstance(node, (Query, And))
            assert all(isinstance(q, Query) for q in node.nodes)
            result.append(
                {q.key: q.value for q in t.cast(t.Sequence[Query], node.nodes)}
            )
        return result

    def _flatten(self) -> Node:
        raise NotImplementedError

    def _and(self, other: Node) -> Node:
        raise NotImplementedError

    def __and__(self, other: Node) -> Node:
        return And(self, other)

    def __or__(self, other: Node) -> Node:
        return Or(self, other)


class And(Node):
    def __init__(self, *nodes: Node) -> None:
        self._nodes = self._filter_nodes(nodes)

    @property
    def nodes(self) -> t.Sequence[Node]:
        return self._nodes

    def _flatten(self) -> Node:
        final: Node | None = None
        for _node in self.nodes:
            node = _node._flatten()
            if final:
                final = final._and(node)
            else:
                final = node
        return final or Or()

    def _and(self, other: Node) -> Node:
        if isinstance(other, Query):
            return And(*self.nodes, other)
        elif isinstance(other, And):
            return And(*self.nodes, *other.nodes)
        else:  # Or
            return other._and(self)

    def __repr__(self) -> str:
        return "({})".format(" AND ".join(str(n) for n in self.nodes))


class Or(Node):
    def __init__(self, *nodes: Node) -> None:
        self._nodes = self._filter_nodes(nodes)

    @property
    def nodes(self) -> t.Sequence[Node]:
        return self._nodes

    def _flatten(self) -> Or:
        return Or(*(n._flatten() for n in self.nodes))

    def _and(self, other: Node) -> Or:
        return Or(*(n._and(other) for n in self.nodes))

    def __repr__(self) -> str:
        return "({})".format(" OR ".join(str(n) for n in self.nodes))


class Query(Node):
    def __init__(self, key: str, value: object) -> None:
        self.key = key
        self.value = value

    @property
    def nodes(self) -> t.Sequence[Node]:
        return (self,)

    def _flatten(self) -> Node:
        return self

    def _and(self, other: Node) -> Node:
        if isinstance(other, Query):
            return And(self, other)
        return other._and(self)

    def __repr__(self) -> str:
        return f"{self.key}={self.value}"



from __future__ import annotations

import typing as t

from detaorm import update
from detaorm.query import Query

if t.TYPE_CHECKING:
    from detaorm.base import Base

__all__ = ("Field",)

_SELF = t.TypeVar("_SELF", bound="Field[t.Any]")
_TYPE = t.TypeVar("_TYPE")


class _Op:
    def __init__(self, op: str | None) -> None:
        self.op = f"?{op}" if op else ""

    def __get__(
        self, inst: object, cls: type[object]
    ) -> t.Callable[[object], Query]:
        def op_func(other: object) -> Query:
            assert isinstance(inst, Field)
            return Query(inst.qual_name + self.op, other)

        return op_func


class Field(t.Generic[_TYPE]):
    """Represent a single field on a Base."""

    name: str
    """The name of the field."""

    @property
    def qual_name(self) -> str:
        """The qualified name of this field."""
        return self.__qual_name

    @qual_name.setter
    def qual_name(self, val: str) -> None:
        self.__qual_name = val

        for k, v in self.__class__.__dict__.items():
            if isinstance(v, Field):
                v.name = k
                v.qual_name = f"{self.qual_name}.{k}"

    def __init_subclass__(cls) -> None:
        for k, v in cls.__dict__.items():
            if isinstance(v, Field):
                v.name = k

    @t.overload
    def __get__(self: _SELF, inst: None, cls: t.Any) -> _SELF:
        ...

    @t.overload
    def __get__(self, inst: Base | NestedField, cls: t.Any) -> _TYPE:
        ...

    def __get__(
        self: _SELF, inst: Base | NestedField | None, cls: t.Any
    ) -> _SELF | _TYPE:
        if isinstance(inst, NestedField) and not isinstance(
            inst, NestedFieldInst
        ):
            return self

        if inst:
            return t.cast(_TYPE, inst.raw[self.name])
        return self

    # query ops
    eq = _Op(None)
    neq = _Op("ne")
    gt = _Op("gt")
    lt = _Op("lt")
    lte = _Op("lte")
    gte = _Op("gte")
    prefix = _Op("pfx")
    range = _Op("range")
    contains = _Op("contains")
    not_contains = _Op("not_contains")

    # updates
    def set(self, value: object) -> update.SetUpdate:
        return update.SetUpdate({self.name: value})

    def increment(self, value: int) -> update.IncrementUpdate:
        return update.IncrementUpdate({self.name: value})

    def append(self, values: list[object]) -> update.AppendUpdate:
        return update.AppendUpdate({self.name: values})

    def prepend(self, values: list[object]) -> update.PrependUpdate:
        return update.PrependUpdate({self.name: values})

    def delete(self) -> update.DeleteUpdate:
        return update.DeleteUpdate([self.name])

    # magic
    __eq__ = eq  # type: ignore
    __ne__ = neq  # type: ignore
    __gt__ = gt
    __lt__ = lt
    __ge__ = gte
    __le__ = lte


class NestedField(Field[t.Dict[str, object]]):
    @property
    def raw(self) -> dict[str, object]:
        raise NotImplementedError

    def __get__(  # type: ignore
        self: _SELF, inst: None | Base | NestedField, cls: t.Any
    ) -> _SELF:
        if isinstance(inst, NestedField) and not isinstance(
            inst, NestedFieldInst
        ):
            return self

        if inst:
            return NestedFieldInst(self, inst)  # type: ignore
        return self


class NestedFieldInst:
    def __init__(self, field: NestedField, base: Base) -> None:
        self.field = field
        self.base = base

    @property
    def raw(self) -> dict[str, object]:
        return self.base.raw[self.field.name]  # type: ignore

    def __getattr__(self, key: str) -> object:
        return getattr(self.field, key).__get__(self, type(self))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['detaorm']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0']

setup_kwargs = {
    'name': 'detaorm',
    'version': '0.3.0',
    'description': 'An async ORM for DetaBase.',
    'long_description': '# DetaORM\nAn async ORM for [DetaBase](https://docs.deta.sh/docs/base/about/).\n\n[Support](https://discord.gg/dGAzZDaTS9) | [PyPI](https://pypi.org/project/detaorm)\n\n## Example Usage\nHere\'s some examples of DetaORM, with commentary. I\'ll add real documentation at some point.\n\n```py\nfrom detaorm import Client, Base, Field\n\n\n# create your base(s) (or models)\nclass User(Base, name="users"):\n    # name is the name of your base. If left empty, it will\n    # default to the lowered name of the class ("user" in\n    # this case.)\n\n    # all bases have a `key` field already.\n\n    # typehints are optional. You could write this instead:\n    # username = Field()\n    username: Field[str] = Field()\n    nicknames: Field[list[str]] = Field()\n\n\n# create the client\n# you have to list the bases when creating the client to\n# allow the client to setup properly.\nclient = Client("<project key>", bases=[User])\n# If you don\'t specify a project key, DetaORM will try to\n# get a project key from the environment, which means that\n# you don\'t need to specify a project keys when running\n# your app with Deta Micros.\n\n# actual usage:\nnew_user = User(username="CircuitSacul")\n\n# all fields are optional, but they will raise a KeyError\n# if you try to access them.\n# The following line will raise a KeyError\nnew_user.nicknames\n\n# to actually create the user, you have to call .insert():\ncreated_user = await new_user.insert()\n# created_user and new_user will be identical\n\n# to update an item:\nupdated_user = await new_user.update(\n    User.nicknames.set(["circuit"])\n)\nprint(updated_user)\n# > User({"key": ..., "username": "CircuitSacul", "nicknames": ["Awesome Person"]})\nupdated_user = await updated_user.update(\n    User.nicknames.append(["sacul"])\n)\nprint(updated_user)\n# > User({"key": ..., "username": "CircuitSacul", "nicknames": ["circuit", "sacul"]})\n\n# updated_user and new_user will be different now.\n# DetaORM sends the query to DetaBase, but also tries\n# to determine the updated value and returns a new item.\nprint(new_user)\n# > User({"key": ..., "username": "CircuitSacul"]})\n\n# you can also use Base.insert_many to insert up to 25 items:\nawait User.insert_many([\n    User(username="BatMan", nicknames=["superhero"]),\n    User(username="SuperMan", nicknames=["superhero"]),\n])\n\n# The easiest way to query items is with .where():\nasync for page in await User.where(User.nicknames.contains("superhero")):\n    for user in page.items:\n        print(user)\n# > User({"key": ..., "username": "BatMan", "nicknames": ["superhero"]})\n# > User({"key": ..., "username": "SuperMan", "nicknames": ["superhero"]})\n```\n',
    'author': 'CircuitSacul',
    'author_email': 'circuitsacul@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import typing as t

from detaorm.types import RAW_ITEM

if t.TYPE_CHECKING:
    from detaorm.client import Client

__all__ = ("RawPaginator", "RawPage")


class RawPaginator:
    def __init__(self, first_page: t.Awaitable[RawPage]) -> None:
        self.__first_page: t.Awaitable[RawPage] | None = first_page
        self.__current_page: RawPage | None = None

    def __aiter__(self) -> RawPaginator:
        return self

    def __await__(self) -> t.Generator[t.Any, None, RawPage]:
        assert self.__first_page
        return self.__first_page.__await__()

    async def __anext__(self) -> RawPage:
        if self.__first_page:
            self.__current_page = await self.__first_page
            self.__first_page = None
            return self.__current_page

        assert self.__current_page
        self.__current_page = await self.__current_page.next()
        if self.__current_page is None:
            raise StopAsyncIteration
        return self.__current_page


class RawPage:
    def __init__(
        self,
        client: Client,
        base_name: str,
        query: t.Sequence[t.Mapping[str, object]] | None,
        limit: int,
        size: int,
        last: str | None,
        items: t.Sequence[RAW_ITEM],
    ) -> None:
        self._base_name = base_name
        self._query = query
        self._client = client
        self._limit = limit
        self._last = last
        self.size = size
        self.items = items

    async def next(self, limit: int | None = None) -> RawPage | None:
        if not self._last:
            return None

        return await self._client.query_items(
            self._base_name,
            self._query,
            limit=limit if limit is not None else self._limit,
            last=self._last,
        )



from __future__ import annotations

import typing as t
from datetime import datetime, timedelta

from detaorm.field import Field

if t.TYPE_CHECKING:
    from detaorm.client import Client, RawPutItemsResponse
    from detaorm.paginator import RawPage, RawPaginator
    from detaorm.query import Node
    from detaorm.update import Update

__all__ = ("Base", "PutItemsResponse")

_BASE = t.TypeVar("_BASE", bound="Base")


class PutItemsResponse(t.Generic[_BASE]):
    def __init__(
        self, response: RawPutItemsResponse, model: type[_BASE]
    ) -> None:
        self.model = model
        self.raw = response

    @property
    def processed(self) -> list[_BASE]:
        return [self.model(**d) for d in self.raw.processed]

    @property
    def failed(self) -> list[_BASE]:
        return [self.model(**d) for d in self.raw.failed]


class Paginator(t.Generic[_BASE]):
    def __init__(self, raw: RawPaginator, model: type[_BASE]) -> None:
        self.model = model
        self.raw = raw

    def __await__(self) -> t.Generator[t.Any, None, Page[_BASE]]:
        async def await_page() -> Page[_BASE]:
            return Page(await self.raw, self.model)

        return await_page().__await__()

    def __aiter__(self) -> Paginator[_BASE]:
        return self

    async def __anext__(self) -> Page[_BASE]:
        return Page(await self.raw.__anext__(), self.model)


class Page(t.Generic[_BASE]):
    def __init__(self, raw: RawPage, model: type[_BASE]) -> None:
        self.model = model
        self.raw = raw

    @property
    def items(self) -> list[_BASE]:
        return [self.model(**d) for d in self.raw.items]

    @property
    def size(self) -> int:
        return self.raw.size

    async def next(self, limit: int | None = None) -> Page[_BASE] | None:
        next = await self.raw.next(limit=limit)
        if next:
            return Page(next, self.model)
        return None


class Base:
    """A Deta "Base", or a single model.

    Args:
        client (Client): The DetaORM Client to use.
        **values (object): Key-value pairs for this model.
    """

    __base_name__: str
    """The name of this Base."""
    _client: Client

    key: Field[str] = Field()

    def __init_subclass__(cls, name: str | None = None) -> None:
        cls.__base_name__ = name or cls.__name__.lower()

        cls.key.name = "key"
        for key, value in cls.__dict__.items():
            if isinstance(value, Field):
                value.name = key
                value.qual_name = key

    def __init__(self, **values: object) -> None:
        self.raw = values
        self._updates: dict[str, object] = {}

    # model methods
    # these methods further abstract the methods on Base.
    async def delete(self) -> None:
        return await self.delete_item(self.key)

    async def insert(
        self: _BASE,
        *,
        expire_at: int | datetime | None = None,
        expire_in: int | timedelta | None = None,
    ) -> _BASE:
        return await self.insert_item(
            self, expire_at=expire_at, expire_in=expire_in
        )

    async def update(
        self: _BASE,
        *updates: Update,
        set: dict[str, object] | None = None,
        increment: dict[str, int] | None = None,
        append: dict[str, list[object]] | None = None,
        prepend: dict[str, list[object]] | None = None,
        delete: list[str] | None = None,
    ) -> _BASE:
        ud = await self.update_item(
            self.key,
            *updates,
            set=set,
            increment=increment,
            append=append,
            prepend=prepend,
            delete=delete,
        )
        new_self = type(self)()
        new_self.raw = self.raw

        def cd(obj: object) -> dict[str, object]:
            assert isinstance(obj, dict)
            assert all(isinstance(k, str) for k in obj.keys())
            return obj

        def cl(obj: object) -> list[str]:
            assert isinstance(obj, list)
            assert all(isinstance(k, str) for k in obj)
            return obj

        for k, v in cd(ud["set"]).items():
            new_self.raw[k] = v
        for k, v in cd(ud["increment"]).items():
            assert isinstance(v, int)
            orig = new_self.raw[k]
            assert isinstance(orig, int)
            new_self.raw[k] = orig + v
        for k, v in cd(ud["append"]).items():
            assert isinstance(v, list)
            orig = new_self.raw[k]
            assert isinstance(orig, list)
            new_self.raw[k] = orig + v
        for k, v in cd(ud["prepend"]).items():
            assert isinstance(v, list)
            orig = new_self.raw[k]
            assert isinstance(orig, list)
            new_self.raw[k] = v + orig
        for k in cl(ud["delete"]):
            new_self.raw.pop(k)

        return new_self

    # abstracted api methods
    # these methods abstract the methods on Client.
    @classmethod
    async def get(cls: type[_BASE], key: str) -> _BASE:
        return cls(**await cls._client.get_item(cls.__base_name__, key))

    @classmethod
    def where(
        cls: type[_BASE],
        query: Node | t.Sequence[t.Mapping[str, object]] | None = None,
        limit: int = 0,
        last: str | None = None,
    ) -> Paginator[_BASE]:
        return Paginator(
            cls._client.query_items(
                cls.__base_name__, query=query, limit=limit, last=last
            ),
            cls,
        )

    # these methods abstract the methods on Client, but are abstracted
    # further on this model.
    @classmethod
    async def insert_many(
        cls: type[_BASE],
        items: t.Sequence[_BASE],
        *,
        expire_at: int | datetime | None = None,
        expire_in: int | timedelta | None = None,
    ) -> PutItemsResponse[_BASE]:
        return PutItemsResponse(
            await cls._client.put_items(
                cls.__base_name__,
                [i.raw for i in items],
                expire_at=expire_at,
                expire_in=expire_in,
            ),
            cls,
        )

    @classmethod
    async def delete_item(cls, key: str) -> None:
        await cls._client.delete_item(cls.__base_name__, key)

    @classmethod
    async def insert_item(
        cls: type[_BASE],
        item: _BASE,
        *,
        expire_at: int | datetime | None = None,
        expire_in: int | timedelta | None = None,
    ) -> _BASE:
        return cls(
            **await cls._client.insert_item(
                cls.__base_name__,
                item.raw,
                expire_at=expire_at,
                expire_in=expire_in,
            )
        )

    @classmethod
    async def update_item(
        cls,
        key: str,
        *updates: Update,
        set: dict[str, object] | None = None,
        increment: dict[str, int] | None = None,
        append: dict[str, list[object]] | None = None,
        prepend: dict[str, list[object]] | None = None,
        delete: list[str] | None = None,
    ) -> dict[str, object]:
        return await cls._client.update_item(
            cls.__base_name__,
            key,
            *updates,
            set=set,
            increment=increment,
            append=append,
            prepend=prepend,
            delete=delete,
        )

    # magic
    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.raw!r})"
